package edu.ntnu.stud.chaosgame.controller.game;

import edu.ntnu.stud.chaosgame.model.data.Vector2D;
import edu.ntnu.stud.chaosgame.view.ChaosGameObserver;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Random;

/**
 * Class representing the chaos game.
 */
public class ChaosGame {

  /**
   * The canvas for this ChaosGame
   */
  private ChaosCanvas canvas;

  /**
   * The chaos game description.
   */
  private ChaosGameDescription description;

  /**
   * Observers monitoring this ChaosGame.
   */
 private ArrayList<ChaosGameObserver> observers;

  /**
   * The current point in the chaos game.
   */
  private Vector2D currentPoint;

  /**
   * The random number generator for the chaos game.
   */
  private Random random;

  /**
   * Number of transforms.
   */
  private int numOfTransforms;

  /**
   * Boolean informing the game whether to use colors for more frequent
   * pixels or not.
   */
  private boolean useColor = false;

  /**
   * Basic parameterised constructor.
   *
   * @param description the ChaosGameDescription.
   * @param canvas the ChaosCanvas.
   */
  public ChaosGame(ChaosGameDescription description, ChaosCanvas canvas) {
    this.canvas = canvas;
    this.description = description;
    this.random = new Random();
    this.numOfTransforms = description.getTransforms().size();
    this.currentPoint = new Vector2D(0, 0);
  }

  /**
   * Parameterised constructor with observers.
   *
   * @param description the ChaosGameDescription.
   * @param canvas the ChaosCanvas.
   * @param observers the ChaosGameObservers to be added.
   */
  public ChaosGame(ChaosGameDescription description, ChaosCanvas canvas, ChaosGameObserver... observers) {
    this.canvas = canvas;
    this.description = description;
    this.random = new Random();
    this.numOfTransforms = description.getTransforms().size();
    this.currentPoint = new Vector2D(0, 0);
    this.observers.addAll(Arrays.asList(observers));
  }


  /**
   * Get the canvas of this chaos game.
   *
   * @return the canvas.
   */
  public ChaosCanvas getCanvas() { return this.canvas; }

  /**
   * Get the current point in the chaos game.
   *
   * @return the point.
   */
  public Vector2D getCurrentPoint() { return this.currentPoint; }

  /**
   * Get the random number generator for this chaos game.
   *
   * @return the RNG.
   */
  public Random getRandom() { return this.random; }

  /**
   * Get the description for this choas game.
   *
   * @return the description.
   */
  public ChaosGameDescription getDescription() {
    return this.description;
  }

  /**
   * Get the chaos game description.
   *
   * @param description the description to set.
   */
  public void setDescription(ChaosGameDescription description) {
    this.description = description;
  }

  /**
   * Set the current point from which the chaos game will proceed to iterate over
   * the set of affine transforms in the chaos game description.
   *
   * @param point the new point.
   */
  public void setCurrentPoint(Vector2D point) {
    this.currentPoint = point;
  }

  /**
   * Set the useColor boolean.
   *
   * @param useColor the boolean to set.
   */
  public void setUseColor(boolean useColor) {
    this.useColor = useColor;
  }


  /**
   * Notify the observers that a change has occurred in the ChaosGame.
   */
  public void notifyObservers() {
    for (ChaosGameObserver observer : this.observers) {
      observer.updateGame(this);
    }
  }

  /**
   * Add an observer to the ChaosGame.
   *
   * @param observer the observer to add.
   */
  public void addObserver(ChaosGameObserver observer) {
    this.observers.add(observer);
  }

  /**
   * Run the chaos game for n iterations.
   *
   * @param n the number of iterations.
   */
  public void runSteps(int n) {
    ArrayList<Integer> results = new ArrayList<>(
        Collections.nCopies(this.description.getTransforms().size(), 0));

    // Testing the RNG
    for (int i = 0; i < n; i++) {

      int j = this.random.nextInt( this.numOfTransforms);

      //System.out.println(j);
      //System.out.println(this.description.getTransforms().size());

      //System.out.println("Before transform: " + currentPoint.getX0() + " " + currentPoint.getX1());
      Vector2D newPoint = this.description.getTransforms().get(j).transform(currentPoint);

      this.currentPoint = this.findValidPoint(newPoint);
      this.canvas.putPixel(currentPoint, useColor);
    }

  }

  /**
   * If the point is out of bounds, iterate until it is within bounds. Used
   * if the useColor boolean is set to false.
   *
   * @param point the starting point.
   * @return the resulting valid point within bounds.
   */
  public Vector2D findValidPoint(Vector2D point) {
    if (!this.canvas.isPointInCanvasRange(point) || this.canvas.getPixel(point) == 1) {
      while (!this.canvas.isPointInCanvasRange(point)) {
        int j = this.random.nextInt(this.numOfTransforms);
        System.out.println("Before transform: " + point.getX0() + " " + point.getX1()); // Test
        point = this.description.getTransforms().get(j).transform(currentPoint);
      }
    }
    return point;
  }

}
