package edu.ntnu.stud.chaosgame.view;

import edu.ntnu.stud.chaosgame.controller.PopupManager;
import javafx.scene.image.ImageView;
import javafx.scene.input.ScrollEvent;
import javafx.scene.transform.Affine;

/**
 * This class extends ImageView to implement proper zooming and panning
 * according to the requirements of the chaos game.
 */
public class ChaosGameImageView extends ImageView {

  /**
   * The controller for this class: a chaos game GUI view.
   */
  private final ChaosGameGui controller;

  /**
   * Affine initialised to the identity matrix.
   */
  private final Affine transform = new Affine();

  private double centreX;
  private double centreY;

  /**
   * The starting x and y positions upon a mouse event.
   */
  private double startX;
  private double startY;

  /**
   * The factor representing the magnitude of the current zoom.
   */
  private double zoomFactor = 1.0;

  /**
   * Integer representing how many recursive levels of QuadTrees
   * are required to get to the current zoom magnitude.
   */
  private int zoomLevel = 0;



  /**
   * Constructor for the ChaosGameImageView.
   *
   * @param controller the GUI which controls this image view.
   */
  public ChaosGameImageView(ChaosGameGui controller) {
    this.setOnScroll(this::userZoom);
    this.setOnMousePressed(this::mousePressed);
    this.setOnMouseDragged(this::mouseDragged);
    //this.setOnMouseReleased(this::mouseReleased);
    this.getTransforms().add(transform);
    this.controller = controller;
    //this.lastCentreX = (float) controller.getWidth() / 2;
    //this.lastCentreY = (float) -controller.getHeight() / 2;

    //this.setStyle("-fx-background-color: white;");
  }

  /**
   * Get the current zoom level.
   *
   * @return the integer representing the zoom level.
   */
  public int getZoomLevel() {return this.zoomLevel; }

  /**
   * Zooms the image view in or out based on the scroll event.
   *
   * @param event the event.
   */
  private synchronized void userZoom(ScrollEvent event) {
    double newZoomFactor = event.getDeltaY() > 0 ? 1.10 : 1 / 1.05;
    try {
      // Get the old values
      double oldScaleX = transform.getMxx();
      double oldScaleY = transform.getMyy();
      double oldTranslateX = transform.getTx();
      double oldTranslateY = transform.getTy();

      // Compute the new values
      double newScaleX = oldScaleX * newZoomFactor;
      double newScaleY = oldScaleY * newZoomFactor;
      double newTranslateX = oldTranslateX - (event.getX() * (newScaleX - oldScaleX));
      double newTranslateY = oldTranslateY - (event.getY() * (newScaleY - oldScaleY));

      // Update the transform
      transform.setMxx(newScaleX);
      transform.setMyy(newScaleY);
      transform.setTx(newTranslateX);
      transform.setTy(newTranslateY);

      this.zoomFactor *= newZoomFactor;
      this.zoomLevel = (int) (Math.log(this.zoomFactor) / Math.log(4)); // Update zoom level.


    } catch (Exception e) {
      PopupManager.displayError("Zoom error", e.getMessage());
    }
  }

  public void fixedZoom(double newZoomFactor) {
    try {
      // Get the old values
      double oldScaleX = transform.getMxx();
      double oldScaleY = transform.getMyy();
      double oldTranslateX = transform.getTx();
      double oldTranslateY = transform.getTy();

      // Compute the new values
      double newScaleX = oldScaleX * newZoomFactor;
      double newScaleY = oldScaleY * newZoomFactor;
      double newTranslateX = oldTranslateX - (this.getWidth() / 2 * (newScaleX - oldScaleX));
      double newTranslateY = oldTranslateY - (this.getHeight() / 2 * (newScaleY - oldScaleY));

      // Update the transform
      transform.setMxx(newScaleX);
      transform.setMyy(newScaleY);
      transform.setTx(newTranslateX);
      transform.setTy(newTranslateY);

      this.zoomFactor *= newZoomFactor;
      this.zoomLevel = (int) (Math.log(this.zoomFactor) / Math.log(4)); // Update zoom level.

    } catch (Exception e) {
      PopupManager.displayError("Zoom error", e.getMessage());
    }
  }
// TODO: remove if unused
  private void updateController() {
    this.controller.updateDetail(this.zoomLevel, this.centreX, this.centreY);
  }

  /**
   * Gets mouse cursor position data when mouse button is pressed.
   *
   * @param event the mouse event.
   */
  private synchronized void mousePressed(javafx.scene.input.MouseEvent event) {
    startX = event.getX();
    startY = event.getY();
  }

  /**
   * Drags the image view based on the mouse cursor position.
   *
   * @param event the mouse event.
   */
  private synchronized void mouseDragged(javafx.scene.input.MouseEvent event) {
    double deltaX = event.getX() - startX;
    double deltaY = event.getY() - startY;

    transform.setTx(transform.getTx() + deltaX);
    transform.setTy(transform.getTy() + deltaY);
  }

  public int getWidth() {
    return (int) this.getImage().getWidth();
  }

  public int getHeight() {
    return (int) this.getImage().getHeight();
  }

  public Affine getTransform() {return this.transform; }

}
